"use strict";

require("core-js/modules/es.array.flat-map.js");

require("core-js/modules/es.array.unscopables.flat-map.js");

exports.__esModule = true;
exports.default = storiesLoader;

var _path = _interopRequireDefault(require("path"));

var _lodash = require("lodash");

var _loaderUtils = _interopRequireDefault(require("loader-utils"));

var _estreeWalker = require("estree-walker");

var _astTypes = require("ast-types");

var _escodegen = require("escodegen");

var _toAst = _interopRequireDefault(require("to-ast"));

var _stripIndent = _interopRequireDefault(require("strip-indent"));

var _getAst = _interopRequireDefault(require("./utils/getAst"));

var _getNameFromFilePath = _interopRequireDefault(require("./utils/getNameFromFilePath"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const getLocalName = index => `__stories_import_${index}`; // Ensure that we have a semicolon (;) at the end: we must put a semicolon
// in front of JSX (`import 'foo'; <Button/>`), otherwise it won't compile


const ensureSemicolon = s => s.replace(/(?:;\s*)?$/, ';'); // Returns the indentation (actual whitespace, not a number of spaces)
// of the first line of code example


const getIndent = (code, position) => code.substring(0, position).match(/[ \t]*$/); // TODO: Deduplicate?


const getModulePathToModuleMap = modules => {
  const moduleMap = {};
  modules.forEach((module, index) => {
    moduleMap[module] = {
      toAST: () => _astTypes.builders.identifier(getLocalName(index))
    };
  });
  return moduleMap;
};

const getImports = ast => {
  const imports = [];
  (0, _estreeWalker.walk)(ast, {
    enter: node => {
      if (node.type === 'ImportDeclaration') {
        var _importDeclarationNod;

        const importDeclarationNode = node;

        if (typeof (importDeclarationNode == null ? void 0 : (_importDeclarationNod = importDeclarationNode.source) == null ? void 0 : _importDeclarationNod.value) === 'string') {
          imports.push(importDeclarationNode.source.value);
        }
      }
    }
  });
  return imports;
};

const getImportDeclarationName = specifier => {
  switch (specifier.type) {
    case 'ImportSpecifier':
    case 'ImportDefaultSpecifier':
    case 'ImportNamespaceSpecifier':
      return specifier.local.name;

    default:
      return '';
  }
};

const getImportStatements = (ast, code) => {
  const imports = [];
  (0, _estreeWalker.walk)(ast, {
    enter: node => {
      if (node.type === 'ImportDeclaration') {
        const importDeclarationNode = node;

        if (importDeclarationNode.source) {
          // @ts-expect-error: There are no types for location ;-/
          const {
            start,
            end
          } = importDeclarationNode;
          imports.push({
            names: importDeclarationNode.specifiers.map(getImportDeclarationName),
            code: code.substring(start, end)
          });
        }
      }
    }
  });
  return imports;
};

const getVariableDeclarationNames = ({
  id
}) => {
  switch (id.type) {
    case 'Identifier':
      return [id.name];

    case 'ObjectPattern':
      return id.properties.map(property => {
        switch (property.type) {
          case 'Property':
            switch (property.value.type) {
              case 'Identifier':
                return property.value.name;
            }

            return '';

          case 'RestElement':
            switch (property.argument.type) {
              case 'Identifier':
                return property.argument.name;
            }

            return '';
        }

        return '';
      }).filter(Boolean);

    default:
      return [];
  }
};

const getVariableStatements = (ast, code) => {
  const variables = [];
  (0, _estreeWalker.walk)(ast, {
    enter: (node, parent) => {
      if (node.type === 'VariableDeclaration' && parent.type === 'Program') {
        const variableDeclarationNode = node; // @ts-expect-error: There are no types for location ;-/

        const {
          start,
          end
        } = variableDeclarationNode;
        variables.push({
          names: (0, _lodash.flatMap)(variableDeclarationNode.declarations, getVariableDeclarationNames),
          code: code.substring(start, end)
        });
      }
    }
  });
  return variables;
};

const getExportCode = (node, code) => {
  switch (node.type) {
    case 'VariableDeclaration':
      {
        const variableDeclarationNode = node;
        const [declaration] = variableDeclarationNode.declarations;

        if (!declaration || declaration.type !== 'VariableDeclarator') {
          break;
        }

        const {
          id,
          init
        } = declaration;

        if (id.type !== 'Identifier' || !init) {
          break;
        }

        switch (init.type) {
          case 'ArrowFunctionExpression':
            {
              // @ts-expect-error 2339: There are no types for location ;-/
              const {
                start,
                end
              } = init.body;
              return {
                name: id.name,
                code: (0, _stripIndent.default)(getIndent(code, start) + code.substring(start, end))
              };
            }
        }

        break;
      }
  }

  return undefined;
}; // This is a very simplified approach: it will add unnecessary dependencies
// because we only check whether the dependency local name is present in the
// example code. However, it won't break the code by not including all the used
// dependencies


const prependExampleWithDependencies = (code, dependencies, componentName) => {
  const usedDependencies = dependencies.filter(dependency => // Ignore the current component import it it's the only import
  !(dependency.names.length === 1 && dependency.names[0] === componentName) && // Include imports when any of the names are present in the code
  dependency.names.some(name => code.match(new RegExp(`\\b${name}\\b`))));
  return [usedDependencies.map(dependency => ensureSemicolon(dependency.code)).filter(Boolean).join('\n'), code].filter(Boolean).join('\n\n');
};

const getExports = (ast, code, componentName) => {
  const imports = getImportStatements(ast, code);
  const variables = getVariableStatements(ast, code);
  const exports = {};
  (0, _estreeWalker.walk)(ast, {
    enter: node => {
      // export const foo = ...
      if (node.type === 'ExportNamedDeclaration') {
        const exportNamedDeclaratioNode = node;

        if (exportNamedDeclaratioNode.declaration) {
          const exportCode = getExportCode(exportNamedDeclaratioNode.declaration, code);

          if (exportCode) {
            exports[(0, _lodash.camelCase)(exportCode.name)] = prependExampleWithDependencies(exportCode.code, [...imports, ...variables], componentName);
          }
        }
      }
    }
  });
  return exports;
};
/**
 * 1. Generate export with all modules imported into a Component Story Format
 *    (CSF) file, so they are available in examples when we run them.
 * 2. Generate export with all strories in a CSF files.
 * 3. Generate imports for all modules imported by a CSF file so weback bundles
 *    the modules.
 * 4. Prepend each story example with necessary imports and local variables.
 *
 * import Button from './Button'
 * import Container from './Container'
 * export const basic = () => <Container><Button /></Container>
 *
 * ->
 *
 * import * as __stories_import_0 from './Button'
 * export const __stories_namedExamples = {
 *   basic: 'import Button from './Button'\n\n<Container><Button /></Container>'
 * }
 * export const __stories_storiesScope = {
 *   './Button': __stories_import_0
 * }
 */


function storiesLoader(storiesCode) {
  const callback = this.async() || (() => '');

  const {
    componentPath,
    mdxDocumentPath
  } = _loaderUtils.default.getOptions(this) || {};
  const storiesAst = (0, _getAst.default)(storiesCode);

  if (!storiesAst) {
    return callback(null, 'export const __stories_namedExamples = {}; export const __stories_storiesScope = {};');
  }

  const componentAbsolutePath = _path.default.resolve(_path.default.dirname(mdxDocumentPath), componentPath);

  const componentName = (0, _getNameFromFilePath.default)(componentAbsolutePath);
  const lines = []; // Export named examples

  const exports = getExports(storiesAst, storiesCode, componentName);
  lines.push(`export const __stories_namedExamples = ${(0, _escodegen.generate)((0, _toAst.default)(exports))}`);
  const imports = getImports(storiesAst); // Generate imports so weback bundles the modules

  imports.forEach((module, index) => {
    lines.push(`import * as ${getLocalName(index)} from '${module}'`);
  }); // Export the mapping of imported modules

  const modulesMap = getModulePathToModuleMap(imports);
  lines.push(`export const __stories_storiesScope = ${(0, _escodegen.generate)((0, _toAst.default)(modulesMap))}`);
  return callback(null, lines.join('\n'));
}